# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

import boto3
import botocore
import time
from botocore.config import Config
from botocore.exceptions import ClientError


from foglamp.plugins.north.aws_iot.awsthing import AwsThing
from foglamp.plugins.north.aws_iot.awsmqtt import AwsMqtt
from foglamp.plugins.north.aws_iot.awsmqtt import _LOGGER

class Singleton(type):
    """ A metaclass used to create singleton classes .
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls):
        try:
            del Singleton._instances[cls]
        except KeyError:
            pass

class AwsClient(metaclass=Singleton):
    """ Singleton class for AWS Boto3 to access AWS Services.
    """
    __slots__ = {'billing_group_exist', 'mqtt_client', 'region', 'access_id', 'secret_key', 
                 'name_to_awsthing', 'thing_types', 'attach_billing_group', 'billing_group', '__weakref__'}
    
    def __init__(self):
        """ Constructor method for AwsClient.
            Args:
                attach_billing_group: A boolean indicating to attach 'Billing Group' to created things.
                billing_group_exist: A boolean indicating if 'Billing Group' exists on AWS IoT.
                mqtt_client: A MQTT object representing connection with AWS IoT.
                region: AWS IoT region for which resources are accessed.
                access_id: Access Id for user authorized to access AWS IoT resources.
                secret_key: Secret key for authorized user.
                name_to_awsthing: A dictionary of 'AwsThing' objects that are created/exist on AWS IoT.
                thing_types: A list containing thingType associated with 'Thing'.
        """
        self.attach_billing_group = False
        self.billing_group_exist = True
        self.billing_group = ''
        self.mqtt_client = None
        self.region = None
        self.access_id = None
        self.secret_key = None
        self.name_to_awsthing = dict()
        self.thing_types = list()
    
    def set_config(self, config):
        """ Setter for attributes received in plugin configuration.
            Args: 
                config: Dict containing aws endpoint, certifcate, key and other configuration.
        """
        self.access_id = config['awsAccessId']['value'].strip()
        self.secret_key = config['awsSecretKey']['value'].strip()
        self.attach_billing_group = bool(config['attachBillingGroup'] == 'true')
        self.billing_group = config['billingGroupName']['value'].strip()
        self.set_region(config['awsIotEndpoint']['value'])

    def configure_mqtt(self, config):
        """ Create and configure MQTT connection object.
            Args: 
                config: Dict containing aws endpoint, certifcate, key and other configuration.
        """
        # if re-configure, shutdown existing connection
        if self.mqtt_client:
            self.mqtt_client.disconnect()
            self.mqtt_client = None
        
        self.mqtt_client = AwsMqtt(config)
        
    def create_thing(self, **kwargs):
        """ Method to create a 'thing' using AWS IoT service client.
        Required Keyword Args:
            thingName: The name of the thing to create. Cannot be changed once created.
            thingTypeName: The name of the thing type associated with the new thing.

        Optional Keyword Args:
            attributePayload: The attribute payload, which consists of up to three name/value pairs in a JSON document.
            billingGroupName : The name of the billing group the thing will be added to.

        Returns:
            A boolean that indicates if thing is created/exists.
        """
        # check if billing group is valid
        if not self.billing_group_exist:
            _LOGGER.debug("Billing group not found on AWS IoT. Reconfigure plugin with valid Billing group name")
            return False
        #request aws client 
        try:
            iot_client = boto3.client('iot', config=Config(region_name=self.region), 
                          aws_access_key_id=self.access_id,
                          aws_secret_access_key=self.secret_key)
            
            if self.attach_billing_group:
                if self.billing_group != '':
                    kwargs['billingGroupName'] = self.billing_group
                else:
                    _LOGGER.warning("Empty 'Biliing Group' in plugin configuration. To add billing group update"
                                    "plugin configuration")
            
            #create thingType
            if not kwargs['thingTypeName'] in self.thing_types:
                success = self.create_thing_type(kwargs['thingTypeName'])
                if not success:
                    return False
                             
            #create IoT Thing
            iot_client.create_thing(**kwargs)
        except ClientError as err:
            _LOGGER.debug("AWS IoT Client exception response: {}".format(err.response))
            errCode = err.response['Error']['Code']
            errMsg = err.response['Error']['Message']
            if errCode == 'ResourceAlreadyExistsException':
                _LOGGER.warning("Thing '{}' already exists on AWS IoT".format(errMsg))
                _LOGGER.info("Updating existing Thing '{}' with properties '{}'".format(kwargs['thingTypeName'], kwargs))
                # Update dict with existing thing
                self.update_thing(**kwargs)
                new_thing = AwsThing()
                new_thing.set_properties(**kwargs)
                self.name_to_awsthing[kwargs['thingName']] = new_thing
                return True
            if errCode == 'ResourceNotFoundException':
                if 'BillingGroup' in errMsg:
                    # BillingGroup should exist prior to running plugin
                    self.billing_group_exist = False
                    _LOGGER.error("Billing group not found on AWS IoT. Reconfigure plugin with valid Billing group name")
            else:
                _LOGGER.exception("Exception occured with AWS IoT service, Error Code '{}',"
                               "Error Message '{}'".format(errCode, errMsg))
            return False
        except botocore.exceptions.ParamValidationError as error:
            _LOGGER.error("The parameters provided are incorrect: {}".format(error))
            return False
        else:
            # Add created thing to dict
            new_thing = AwsThing()
            new_thing.set_properties(**kwargs)
            self.name_to_awsthing[kwargs['thingName']] = new_thing
            _LOGGER.info("AWS IoT Thing '{}' created successfully with following properties {}".format(kwargs['thingName'], kwargs))
            return True

    def update_thing(self, **kwargs):
        """ Method to update existing 'Thing' using AWS IoT service client.
        Required Keyword Args:
            thingName: The name of the thing to create. Cannot be changed once created.

        Optional Keyword Args:
            thingTypeName: The name of the thing type associated with the new thing.
            attributePayload: The attribute payload, which consists of up to three name/value pairs in a JSON document.

        Returns:
            A boolean that indicates if thing is updated.
        """
        try:
            iot_client = boto3.client('iot', config=Config(region_name=self.region), 
                          aws_access_key_id=self.access_id,
                          aws_secret_access_key=self.secret_key)
            # Merge always true for attributes
            if kwargs['attributePayload'] is not None:
                    kwargs['attributePayload']['merge'] = True
            
            # check if thingType changed
            if not kwargs['thingTypeName'] in self.thing_types:
                success = self.create_thing_type(kwargs['thingTypeName'])
                if not success:
                    _LOGGER.warning("Unable to update AWS IoT Thing '{}' with ThingType '{}'"
                                    .format(kwargs['thingName'], kwargs['thingTypeName']))
                    return False
                
            #update thing
            iot_client.update_thing(**kwargs)
            _LOGGER.info("AWS IoT Thing '{}' updated successfully with following properties {}".format(kwargs['thingName'], kwargs))
        except ClientError as err:
            _LOGGER.debug("AWS IoT Client exception while updating Thing: {}".format(err.response))
            errCode = err.response['Error']['Code']
            errMsg = err.response['Error']['Message']
            _LOGGER.exception("Exception occured while updating thing '{}', Error Code '{}',"
                               "Error Message '{}'".format(kwargs['thingName'], errCode, errMsg))
            
            return False
        
        return True
    
    def create_thing_type(self, type_name):
        """ Method to create a 'thingType' using AWS IoT service client.
        Args:
            type_name: Name of Thing Type.

        Returns:
            A boolean that indicates if Thing Type is created/exists.
        """
        try:
            iot_client = boto3.client('iot', config=Config(region_name=self.region), 
                          aws_access_key_id=self.access_id,
                          aws_secret_access_key=self.secret_key)
            
            iot_client.create_thing_type(thingTypeName=type_name)
        except ClientError as err:
            _LOGGER.debug("AWS IoT Client exception while creating thing '{}'".format(err.response))
            errCode = err.response['Error']['Code']
            errMsg = err.response['Error']['Message']
            if errCode == 'ResourceAlreadyExistsException':
                # add to thingType list
                self.thing_types.append(type_name)
                return True
            else:
                _LOGGER.exception("Exception occured at AWS IoT service Error Code '{}', "
                               "Error Message '{}'".format(errCode, errMsg))
        except botocore.exceptions.ParamValidationError as error:
            _LOGGER.exception('The parameters provided for AWS IoT client are incorrect: {}. Check AWSHint for'
                              ' valid values'.format(error))
        else:
            self.thing_types.append(type_name)
            _LOGGER.info("AWS IoT ThingType '{}' created successfully.".format(type_name))
            return True
        
        return False

    def thing_exists(self, name):
        """ Helper method to verify if 'thing' exists on AWS IoT.
        Args:
            name: Name of thing.

        Returns:
            A boolean that indicates if thing exists.
        """
        if len(self.name_to_awsthing) > 0:
            if name in self.name_to_awsthing:
                return True
        return False

    def set_region(self, endpoint):
        """ Method to extract region name from 'endpoint' given in plugin configuration.
        Args:
            endpoint: Host name of AWS IoT server.
        """
        try:
            start_pattern = '.iot.'
            end_pattern = '.amazonaws.com'
            start_index = endpoint.index(start_pattern)
            end_index = endpoint.index(end_pattern)
            self.region = endpoint[start_index+len(start_pattern):end_index]
        except:
            _LOGGER.error("Invalid AWS endpoint URL, unable to start the plugin."
                          "Re-configure with valid AWS endpoint URL")
            
    def clean(self):
        if self.mqtt_client:
            self.mqtt_client.shutdown_connection = True
            self.mqtt_client.connect_thread = None
        time.sleep(2)
        del self.mqtt_client
        self.mqtt_client = None
        self.__class__.clear()

